#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>

#define STATS	"/sys/thermal_framework/stats"
#define NUMTEMPS 5
#define MAXFREQS 5
#define BUFSIZE (10*80)

/*
 * Sample of the format of the stats file:

cputime (ms) for celsius temperatures at each frequency                        
                                                                                
Frequency        <  40    40 to  43    43 to  46    46 to  49        >  49      
                                                                                
   350000:        3710            0            0            0            0      
   700000:           0            0            0            0            0      
   920000:           0            0            0            0            0      
  1200000:           0            0            0            0            0      
                                                                                
Total Time: 3.91 s

 */

struct _data {
	int temps[NUMTEMPS - 1];    // upper limit on each; NUMTEMPS-1 
			 	    // boundaries identifies NUMTEMPS buckets
	int freqs[MAXFREQS];
	unsigned long time[MAXFREQS][NUMTEMPS];
	unsigned long totaltime_s;
	unsigned totaltime_ms;
} orgdata, currdata;

int gotsignal = 0;

void usage(char *prog)
{
	printf("Usage: %s -o output [-r secs ] [ -c ]\n", prog);
	printf("   -o output  Location to store the stats.  If output is a\n");
	printf("              directory, the stats will be stored in a file\n");
	printf("              named according to time %s started.\n", prog);
	printf("              Otherwise, specify a file name for output\n");
	printf("   -r secs    The time, in seconds, between stats refreshes\n");
	printf("              default: 10 seconds\n");
	printf("   -c         Cumulative stats are to be maintained in the\n");
	printf("              output file.  %s will add new times to the\n",
					prog);
	printf("              values found in the output file at the time\n");
	printf("              it starts.  This option is only available if\n");
	printf("              the output is a file and not a directory\n\n");
	printf("Notes: When using cumulative stats, if there is ever a\n");
	printf("       mismatch in the temperature ranges between new and\n");
	printf("       old stats, the stats will be reset to 0.\n");

}

void sighandler(int sig)
{
	gotsignal = 1;
}

/*
 * Parse the line with the format:
 *     Frequency      <  40    40 to  43    43 to  46    46 to  49      >  49
 * for the temperature boundaries.
 *
 * returns -1 if not able to parse, 0 when no error
 */
int parsestats(char *buf, struct _data *data)
{
	char *buf1;
	int ret, i = 0;
	int unused1, unused2, unused3;

	if (!buf || !data)
		return -1;
	if ((buf = strstr(buf, "Frequency")) == NULL)
		return -1;
	ret = sscanf(buf, "Frequency < %d  %d to %d  %d to %d  %d to %d",
			&data->temps[0], &unused1, &data->temps[1], &unused2, 
			&data->temps[2], &unused3, &data->temps[3]);
	if (ret != 7)
		goto fail;

	do {
		/* move to the next line */
		if (buf1 = strstr(buf, "\n"))
			buf = buf1 + 1;

		/* skip blank lines and spaces */
		while ((*buf == '\n') || (*buf == '\r') || (*buf == ' '))
			buf++;
		ret = sscanf(buf, "%d: %lu %lu %lu %lu %lu", &data->freqs[i],
				&data->time[i][0], &data->time[i][1],
				&data->time[i][2], &data->time[i][3],
				&data->time[i][4]);
		i++;
	} while ((i<MAXFREQS) && (*buf) && (ret == 6));

	if ((buf = strstr(buf, "Total Time")) == NULL)
		goto fail;

	ret = sscanf(buf, "Total Time: %lu.%u s", &data->totaltime_s,
			&data->totaltime_ms);
	if (ret == 2)
		return i;

fail:
	/* We go start over from zero whenever we can't parse the data */
	memset(data, 0, sizeof(struct _data));
	return -1;
}

int writestats(FILE *file, struct _data *data)
{
	int i, j;


	if (!file)
		return -1;

	fprintf(file, "cputime (ms) for celsius temperatures at each "
			"frequency\n\n");
	fprintf(file, "Frequency        < %3d", data->temps[0]);
	for (j=1; j<NUMTEMPS-1; j++)
		fprintf(file, "   %3d to %3d",
				data->temps[j-1], data->temps[j]);
	fprintf(file, "        > %3d\n\n", data->temps[j-1]);
	for (i = 0; (i < MAXFREQS) && (data->freqs[i] != 0); i++) {
		fprintf(file, "%9u:", data->freqs[i]);
		for (j=0; j<NUMTEMPS; j++) {
			fprintf(file, "%12lu ", data->time[i][j]);
		}
		fprintf(file, "\n");
	}
	fprintf(file, "\nTotal Time: %lu.%u s\n",
				data->totaltime_s, data->totaltime_ms);
}

int main(int argc, char **argv)
{
	int opt, ret, i, j;
	char *outname = NULL;
	FILE *file = NULL;
	FILE *sys = NULL;
	int cumulative = 0, rate = 10;
	int output_is_dir = 0, output_exists = 0;
	int doreset = 0;
	struct stat sb;
	char withdate[100];
	char buff[BUFSIZE];

	while((opt = getopt(argc, argv, "o:r:c")) != -1) {
		switch(opt) {
		case 'o':
			outname = optarg;
			break;
		case 'r':
			rate = atoi(optarg);
			break;
		case 'c':
			cumulative = 1;
			break;
		default:
			usage(argv[0]);
			exit(1);
		}
	}
	if (outname == NULL) {
		fprintf(stderr, "Error: No output specified\n");
		usage(argv[0]);
		exit(1);
	}
	if (rate < 1) {
		fprintf(stderr, "Invalid rate\n");
		exit(1);
	}

	if (stat(STATS, &sb) < 0) {
		fprintf(stderr, "Error: stats source %s is not available:\n"
				"\t%s\n", STATS, strerror(errno));
		if (errno == ENOENT) {
			fprintf(stderr, "The kernel must be compiled with "
					"CONFIG_THERMAL_FRAMEWORK_STATS=y\n");
		}
		exit(1);
	}

	if (stat(outname, &sb) == 0) {
		output_exists = 1;
		if (S_ISDIR(sb.st_mode)) {
			output_is_dir = 1;
		}
	}

	if (output_exists && output_is_dir && cumulative) {
		fprintf(stderr, "For cumulative stats, a file rather than a "
				"directory must be selected\nfor output");
		exit(1);
	}
	if (output_exists && !output_is_dir && cumulative) {
		/* Make a backup copy of the stats file */
		sprintf(buff, "cp %s %s.old\n", outname, outname);
		system(buff);
	}
	if (output_exists && output_is_dir && !cumulative) {
		time_t t;
		struct tm *mytm;

		/* Store stats in a file named with the date+time */
		if (time(&t) && ((mytm = localtime(&t)) != NULL)) {
			snprintf(withdate, 100, "%s/%02d-%02d-%02d.%02d.%02d",
					outname, mytm->tm_year % 100,
					mytm->tm_mon, mytm->tm_mday,
					mytm->tm_hour, mytm->tm_min);
			outname = withdate;
		}
	}


	if (output_exists && cumulative) {
		if ((sys = fopen(STATS, "w")) == NULL) {
			fprintf(stderr, "Unable to open file %s:\n\t%s\n", 
					STATS, strerror(errno));
			exit(1);
		}
		/* Read the original values found in the file */
		if ((file = fopen(outname, "r")) == NULL) {
			fprintf(stderr, "Unable to open file %s:\n"
				"\t%s\n", outname, strerror(errno));
			exit(1);
		}
		if (fread(buff, BUFSIZE, 1, file) < 0) {
			fprintf(stderr, "Unable to read prior values from %s:"
				"\n\t%s\n", outname, strerror(errno));
			exit(1);
		}
		if (parsestats(buff, &orgdata) <= 0) {
			fprintf(stderr, "Unable to parse prior values from "
				"%s.\nStarting from 0.\n", outname);
			doreset = 1;
		}
		fclose(file);
		/*
		 * When doing cumulative, we need to clear the stats in the
		 * kernel.  Either the system was just booted (clearing makes 
		 * little difference) or this program was just restarted.
		 * In the latter case, we want to avoid an overlap in the
		 * stats between what the kernel reports and what was written
		 * in the stats file the last time this program ran.
		 */
		fprintf(sys, "anything");
		fclose(sys);
	}

	/* 
	 * Catch SIGTERM so that we can be cleanly shut down down not
	 * writing to the output file.  All other signals we ignore.
	 */
	for (i=0; i<_NSIG; i++)
		if ((i != SIGTERM) && (i != SIGINT))
			signal(i, SIG_IGN);
	signal(SIGTERM, sighandler);
	signal(SIGINT, sighandler);

	while (!gotsignal)
	{
		if ((sys = fopen(STATS, "r")) == NULL) {
			fprintf(stderr, "Unable to open file %s:\n\t%s\n", 
					STATS, strerror(errno));
		} else if ((file = fopen(outname, "w")) == NULL) {
			fprintf(stderr, "Unable to open file %s:\n\t%s\n", 
					outname, strerror(errno));
		} else if (fread(buff, BUFSIZE, 1, sys) < 0) {
			fprintf(stderr, "Unable to read stats values from %s:"
				"\n\t%s\n", STATS, strerror(errno));
		} else if (parsestats(buff, &currdata) <= 0) {
			fprintf(stderr, "Unable to parse stats\n");
		} else {
			if (cumulative) {
				/* Did the temp ranges change ? */
				for (j=0; j<NUMTEMPS-1; j++) {
					if (currdata.temps[j] != 
							orgdata.temps[j]) {
						/* Need to ignore old data */
						cumulative = 0;
						break;
					}
				}
			}
			if (cumulative) {
				for (i=0; i<MAXFREQS; i++)
					for (j=0; j<NUMTEMPS; j++)
						currdata.time[i][j] +=
							orgdata.time[i][j];
				currdata.totaltime_s += orgdata.totaltime_s +
					(currdata.totaltime_ms + 
						orgdata.totaltime_ms) / 1000;
				currdata.totaltime_ms = (currdata.totaltime_ms 
					+ orgdata.totaltime_ms) % 1000;
			}
			writestats(file, &currdata);
		}
		fclose(file);
		fclose(sys);
		sleep(rate);
	}
}
